@@ -1,9 +1,9 @@
 import inspect
 import re
-from functools import lru_cache
+from functools import lru_cache, reduce
+from operator import mul
 from types import FunctionType
 
-import numba
 import numpy as np
 
 from ... import _STANDARD_OPERATOR_NAMES, backend, binary, monoid, op
@@ -19,12 +19,12 @@ from ...dtypes import (
     UINT16,
     UINT32,
     UINT64,
-    _sample_values,
     _supports_complex,
     lookup_dtype,
 )
 from ...exceptions import UdfParseError, check_status_carg
-from .. import ffi, lib
+from .. import _has_numba, _supports_udfs, ffi, lib
+from ..dtypes import _sample_values
 from ..expr import InfixExprBase
 from .base import (
     _SS_OPERATORS,
@@ -33,16 +33,46 @@ from .base import (
     TypedOpBase,
     _call_op,
     _deserialize_parameterized,
-    _get_udt_wrapper,
     _hasop,
-    _udt_mask,
 )
 
+if _has_numba:
+    import numba
+
+    from .base import _get_udt_wrapper
 if _supports_complex:
     from ...dtypes import FC32, FC64
 
 ffi_new = ffi.new
 
+if _has_numba:
+    _udt_mask_cache = {}
+
+    def _udt_mask(dtype):
+        """Create mask to determine which bytes of UDTs to use for equality check."""
+        if dtype in _udt_mask_cache:
+            return _udt_mask_cache[dtype]
+        if dtype.subdtype is not None:
+            mask = _udt_mask(dtype.subdtype[0])
+            N = reduce(mul, dtype.subdtype[1])
+            rv = np.concatenate([mask] * N)
+        elif dtype.names is not None:
+            prev_offset = mask = None
+            masks = []
+            for name in dtype.names:
+                dtype2, offset = dtype.fields[name]
+                if mask is not None:
+                    masks.append(np.pad(mask, (0, offset - prev_offset - mask.size)))
+                mask = _udt_mask(dtype2)
+                prev_offset = offset
+            masks.append(np.pad(mask, (0, dtype.itemsize - prev_offset - mask.size)))
+            rv = np.concatenate(masks)
+        else:
+            rv = np.ones(dtype.itemsize, dtype=bool)
+        # assert rv.size == dtype.itemsize
+        _udt_mask_cache[dtype] = rv
+        return rv
+
 
 class TypedBuiltinBinaryOp(TypedOpBase):
     __slots__ = ()
@@ -170,7 +200,7 @@ class ParameterizedBinaryOp(ParameterizedUdf):
 
     @property
     def commutes_to(self):
-        if type(self._commutes_to) is str:
+        if isinstance(self._commutes_to, str):
             self._commutes_to = BinaryOp._find(self._commutes_to)
         return self._commutes_to
 
@@ -476,7 +506,7 @@ class BinaryOp(OpBase):
                         type_.gb_obj,
                     ),
                     "BinaryOp",
-                    new_binary,
+                    new_binary[0],
                 )
                 op = TypedUserBinaryOp(new_type_obj, name, type_, ret_type, new_binary[0])
                 new_type_obj._add(op)
@@ -493,8 +523,8 @@ class BinaryOp(OpBase):
         if dtypes in self._udt_types:
             return self._udt_ops[dtypes]
 
-        nt = numba.types
-        if self.name == "eq" and not self._anonymous:
+        if self.name == "eq" and not self._anonymous and _has_numba:
+            nt = numba.types
             # assert dtype.np_type == dtype2.np_type
             itemsize = dtype.np_type.itemsize
             mask = _udt_mask(dtype.np_type)
@@ -531,7 +561,8 @@ class BinaryOp(OpBase):
                     #     z_ptr[0] = True
                     z_ptr[0] = (x[mask] == y[mask]).all()
 
-        elif self.name == "ne" and not self._anonymous:
+        elif self.name == "ne" and not self._anonymous and _has_numba:
+            nt = numba.types
             # assert dtype.np_type == dtype2.np_type
             itemsize = dtype.np_type.itemsize
             mask = _udt_mask(dtype.np_type)
@@ -567,6 +598,8 @@ class BinaryOp(OpBase):
                     #     z_ptr[0] = False
                     z_ptr[0] = (x[mask] != y[mask]).any()
 
+        elif self._numba_func is None:
+            raise KeyError(f"{self.name} does not work with {dtypes} types")
         else:
             numba_func = self._numba_func
             sig = (dtype.numba_type, dtype2.numba_type)
@@ -581,7 +614,7 @@ class BinaryOp(OpBase):
                 new_binary, binary_wrapper.cffi, ret_type._carg, dtype._carg, dtype2._carg
             ),
             "BinaryOp",
-            new_binary,
+            new_binary[0],
         )
         op = TypedUserBinaryOp(
             self,
@@ -600,27 +633,98 @@ class BinaryOp(OpBase):
         """Register a BinaryOp without registering it in the ``graphblas.binary`` namespace.
 
         Because it is not registered in the namespace, the name is optional.
+
+        Parameters
+        ----------
+        func : FunctionType
+            The function to compile. For all current backends, this must be able
+            to be compiled with ``numba.njit``.
+            ``func`` takes two input parameters of any dtype and returns any dtype.
+        name : str, optional
+            The name of the operator. This *does not* show up as ``gb.binary.{name}``.
+        parameterized : bool, default False
+            When True, create a parameterized user-defined operator, which means
+            additional parameters can be "baked into" the operator when used.
+            For example, ``gb.binary.isclose`` is a parameterized function that
+            optionally accepts ``rel_tol`` and ``abs_tol`` parameters, and it
+            can be used as: ``A.ewise_mult(B, gb.binary.isclose(rel_tol=1e-5))``.
+            When creating a parameterized user-defined operator, the ``func``
+            parameter must be a callable that *returns* a function that will
+            then get compiled.
+        is_udt : bool, default False
+            Whether the operator is intended to operate on user-defined types.
+            If True, then the function will not be automatically compiled for
+            builtin types, and it will be compiled "just in time" when used.
+            Setting ``is_udt=True`` is also helpful when the left and right
+            dtypes need to be different.
+
+        Returns
+        -------
+        BinaryOp or ParameterizedBinaryOp
         """
+        cls._check_supports_udf("register_anonymous")
         if parameterized:
             return ParameterizedBinaryOp(name, func, anonymous=True, is_udt=is_udt)
         return cls._build(name, func, anonymous=True, is_udt=is_udt)
 
     @classmethod
     def register_new(cls, name, func, *, parameterized=False, is_udt=False, lazy=False):
-        """Register a BinaryOp. The name will be used to identify the BinaryOp in the
-        ``graphblas.binary`` namespace.
+        """Register a new BinaryOp and save it to ``graphblas.binary`` namespace.
 
-            >>> def max_zero(x, y):
-                    r = 0
-                    if x > r:
-                        r = x
-                    if y > r:
-                        r = y
-                    return r
-            >>> gb.core.operator.BinaryOp.register_new("max_zero", max_zero)
-            >>> dir(gb.binary)
-            [..., 'max_zero', ...]
+        Parameters
+        ----------
+        name : str
+            The name of the operator. This will show up as ``gb.binary.{name}``.
+            The name may contain periods, ".", which will result in nested objects
+            such as ``gb.binary.x.y.z`` for name ``"x.y.z"``.
+        func : FunctionType
+            The function to compile. For all current backends, this must be able
+            to be compiled with ``numba.njit``.
+            ``func`` takes two input parameters of any dtype and returns any dtype.
+        parameterized : bool, default False
+            When True, create a parameterized user-defined operator, which means
+            additional parameters can be "baked into" the operator when used.
+            For example, ``gb.binary.isclose`` is a parameterized function that
+            optionally accepts ``rel_tol`` and ``abs_tol`` parameters, and it
+            can be used as: ``A.ewise_mult(B, gb.binary.isclose(rel_tol=1e-5))``.
+            When creating a parameterized user-defined operator, the ``func``
+            parameter must be a callable that *returns* a function that will
+            then get compiled. See the ``user_isclose`` example below.
+        is_udt : bool, default False
+            Whether the operator is intended to operate on user-defined types.
+            If True, then the function will not be automatically compiled for
+            builtin types, and it will be compiled "just in time" when used.
+            Setting ``is_udt=True`` is also helpful when the left and right
+            dtypes need to be different.
+        lazy : bool, default False
+            If False (the default), then the function will be automatically
+            compiled for builtin types (unless ``is_udt`` is True).
+            Compiling functions can be slow, however, so you may want to
+            delay compilation and only compile when the operator is used,
+            which is done by setting ``lazy=True``.
+
+        Examples
+        --------
+        >>> def max_zero(x, y):
+                r = 0
+                if x > r:
+                    r = x
+                if y > r:
+                    r = y
+                return r
+        >>> gb.core.operator.BinaryOp.register_new("max_zero", max_zero)
+        >>> dir(gb.binary)
+        [..., 'max_zero', ...]
+
+        This is how ``gb.binary.isclose`` is defined:
+
+        >>> def user_isclose(rel_tol=1e-7, abs_tol=0.0):
+        >>>     def inner(x, y):
+        >>>         return x == y or abs(x - y) <= max(rel_tol * max(abs(x), abs(y)), abs_tol)
+        >>>     return inner
+        >>> gb.binary.register_new("user_isclose", user_isclose, parameterized=True)
         """
+        cls._check_supports_udf("register_new")
         module, funcname = cls._remove_nesting(name)
         if lazy:
             module._delayed[funcname] = (
@@ -681,21 +785,22 @@ class BinaryOp(OpBase):
                     orig_op.gb_name,
                 )
                 new_op._add(cur_op)
-        # Add floordiv
-        # cdiv truncates towards 0, while floordiv truncates towards -inf
-        BinaryOp.register_new("floordiv", _floordiv, lazy=True)  # cast to integer
-        BinaryOp.register_new("rfloordiv", _rfloordiv, lazy=True)  # cast to integer
+        if _supports_udfs:
+            # Add floordiv
+            # cdiv truncates towards 0, while floordiv truncates towards -inf
+            BinaryOp.register_new("floordiv", _floordiv, lazy=True)  # cast to integer
+            BinaryOp.register_new("rfloordiv", _rfloordiv, lazy=True)  # cast to integer
 
-        # For aggregators
-        BinaryOp.register_new("absfirst", _absfirst, lazy=True)
-        BinaryOp.register_new("abssecond", _abssecond, lazy=True)
-        BinaryOp.register_new("rpow", _rpow, lazy=True)
+            # For aggregators
+            BinaryOp.register_new("absfirst", _absfirst, lazy=True)
+            BinaryOp.register_new("abssecond", _abssecond, lazy=True)
+            BinaryOp.register_new("rpow", _rpow, lazy=True)
 
-        # For algorithms
-        binary._delayed["binom"] = (_register_binom, {})  # Lazy with custom creation
-        op._delayed["binom"] = binary
+            # For algorithms
+            binary._delayed["binom"] = (_register_binom, {})  # Lazy with custom creation
+            op._delayed["binom"] = binary
 
-        BinaryOp.register_new("isclose", _isclose, parameterized=True)
+            BinaryOp.register_new("isclose", _isclose, parameterized=True)
 
         # Update type information with sane coercion
         position_dtypes = [
@@ -777,14 +882,23 @@ class BinaryOp(OpBase):
             if right_name not in binary._delayed:
                 if right_name in _SS_OPERATORS:
                     right = binary._deprecated[right_name]
-                else:
+                elif _supports_udfs:
                     right = getattr(binary, right_name)
+                else:
+                    right = getattr(binary, right_name, None)
+                    if right is None:
+                        continue
                 if backend == "suitesparse" and left_name in _SS_OPERATORS:
                     right._commutes_to = f"ss.{left_name}"
                 else:
                     right._commutes_to = left_name
         for name in cls._commutative:
-            cur_op = getattr(binary, name)
+            if _supports_udfs:
+                cur_op = getattr(binary, name)
+            else:
+                cur_op = getattr(binary, name, None)
+                if cur_op is None:
+                    continue
             cur_op._commutes_to = name
         for left_name, right_name in cls._commutes_to_in_semiring.items():
             if left_name in _SS_OPERATORS:
@@ -805,7 +919,10 @@ class BinaryOp(OpBase):
             (binary.any, _first),
         ]:
             binop.orig_func = func
-            binop._numba_func = numba.njit(func)
+            if _has_numba:
+                binop._numba_func = numba.njit(func)
+            else:
+                binop._numba_func = None
             binop._udt_types = {}
             binop._udt_ops = {}
         binary.any._numba_func = binary.first._numba_func
