@@ -3,15 +3,19 @@ import itertools
 import numpy as np
 
 from .. import backend, binary, config, monoid
-from ..binary import isclose
-from ..dtypes import _INDEX, FP64, lookup_dtype, unify
+from ..dtypes import _INDEX, FP64, _index_dtypes, lookup_dtype, unify
 from ..exceptions import EmptyObject, check_status
-from . import automethods, ffi, lib, utils
+from . import _has_numba, _supports_udfs, automethods, ffi, lib, utils
 from .base import BaseExpression, BaseType, call
 from .expr import AmbiguousAssignOrExtract
 from .operator import get_typed_op
 from .utils import _Pointer, output_type, wrapdoc
 
+if _supports_udfs:
+    from ..binary import isclose
+else:
+    from .operator.binary import _isclose as isclose
+
 ffi_new = ffi.new
 
 
@@ -154,7 +158,11 @@ class Scalar(BaseType):
     def __complex__(self):
         return complex(self.value)
 
-    __index__ = __int__
+    @property
+    def __index__(self):
+        if self.dtype in _index_dtypes:
+            return self.__int__
+        raise AttributeError("Scalar object only has `__index__` for integral dtypes")
 
     def __array__(self, dtype=None):
         if dtype is None:
@@ -261,6 +269,17 @@ class Scalar(BaseType):
             return False
         # We can't yet call a UDF on a scalar as part of the spec, so let's do it ourselves
         isclose_func = isclose(rel_tol, abs_tol)
+        if not _has_numba:
+            # Check if types are compatible
+            get_typed_op(
+                binary.eq,
+                self.dtype,
+                other.dtype,
+                is_left_scalar=True,
+                is_right_scalar=True,
+                kind="binary",
+            )
+            return isclose_func(self.value, other.value)
         isclose_func = get_typed_op(
             isclose_func,
             self.dtype,
@@ -1041,7 +1060,7 @@ def _as_scalar(scalar, dtype=None, *, is_cscalar):
 
 
 def _dict_to_record(np_type, d):
-    """Converts e.g. `{"x": 1, "y": 2.3}` to `(1, 2.3)`."""
+    """Converts e.g. ``{"x": 1, "y": 2.3}`` to ``(1, 2.3)``."""
     rv = []
     for name, (dtype, _) in np_type.fields.items():
         val = d[name]
